# clock.py
#
#

""" timer, repeater and other clock based classes. """

__copyright__ = "Copyright 2015, Bart Thate"

## IMPORTS

from meds.utils.time import get_day, get_hour, today, short_date
from meds.dispatcher import Dispatcher
from meds.utils.misc import dumps
from meds.utils.name import named
from meds.parsed import Parsed
from meds.object import Object
from meds.kernel import kernel
from meds.store import Store
from meds.event import Event

import threading
import logging
import time

# TIMER

class Timer(Object):

    """ call function on a certain time. """

    def __init__(self, time_alarm, event, *args, **kwargs):
        Object.__init__(self, *args, **kwargs)
        self.sleep = time_alarm - time.time()
        self.event = event

    def start(self, *args, **kwargs):
        """ start the Timer. """
        self._timer = threading.Timer(self.sleep, self.echo)
        self._timer.setDaemon(True)
        self._timer.setName(named(self.echo))
        self._timer._start = time.time()
        self._timer._last = time.time()
        self._timer._status = "start"
        self._timer._status = "ok"
        self._timer.sleep = self.sleep
        logging.warn("# sleeping %s seconds" % self.sleep) 
        self._timer.start()

    def echo(self, *args, **kwargs):
        """ announce on fleet. """
        for bot in kernel.fleet: bot.announce(self.event.txt)
        self.event.done = True
        self.event.save()

    def stop(self, *args, **kwargs):
        """ stop the Timer. """
        self._timer._status = "stopped"
        self._timer.cancel()

running = []
        
## LOOPS

class Repeater(Object):

    """ repeater calls a function every x seconds. """ 

    def __init__(self, sleep, func, *args, **kwargs):
        """ repeat operation at timeinterval. kwargs get run on exec time. """
        Object.__init__(self, **kwargs)
        self.kwargs = kwargs
        self.args = args 
        self.func = func
        self.sleep = sleep
        self._start = time.time()
        self._status = "start"
        self._running = named(self.func)

    def start(self, *args, **kwargs):
        """ initialize a Repeater. Accepts a named "name" argument. """
        self._timer = threading.Timer(self.sleep, self.run)
        self._timer.setDaemon(True)
        if "name" in self.kwargs: self._timer.setName(self.kwargs["name"])
        else: self._timer.setName(named(self.func))
        self._timer._nrs = Object()
        self._timer._nrs.loops = 0
        self._timer._start = time.time()
        self._timer._last = time.time()
        self._timer._status = "ok"
        self._timer.sleep = self.sleep
        self._timer.start()

    def run(self, *args, **kwargs):
        """ run registered function. """
        self.start()
        self._timer._status = "ok"
        self._timer._nrs.loops += 1
        self.func(**self.kwargs)
 
    def exit(self, *args, **kwargs):
        """ stop the Repeater. """
        self._status = ""
        self._timer.cancel()

## COMMANDS

def clock_timer(*args, **kwargs):
    p = Parsed(*args, **kwargs)
    if not p.rest: return
    day = get_day(p.rest) or today()
    hour = get_hour(p.rest)
    target = day + get_hour(p.rest)
    t = Object()
    t.services = "clock"
    t.prefix = "timer"
    t.txt = p.rest
    t.time = target
    t.done = False
    t.save()
    timer = Timer(target, t)
    kernel.start(timer.start)
    p.ok()
    logging.warn("# Timer/start %s" % dumps(t, indent=4))

kernel.register("clock.timer", clock_timer)

## INIT

def init():
    store = Store()
    for fn in store.all("timer"):
        o = Event().load(fn)
        if "time" not in o: continue
        if time.time() < int(o.time):
            timer = Timer(int(o.time), o)
            kernel.start(timer.start)
