# meds/dispatcher.py
#
#

""" working unit of meds. """

__copyright__ = "Copyright 2015, Bart Thate"

## IMPORTS

from meds.errors import NoFunction, NoEvent, RemoteDisconnect
from meds.utils.name import naming, named, typed, naming
from meds.utils.trace import get_plugname, get_strace
from meds.utils.misc import dumps, slice, reg
from meds.utils.parse import txt_parse
from meds.errors import error, NoJSON
from meds.utils.time import elapsed
from meds.object import Object

import collections
import threading
import logging
import random
import queue
import time

## TASKS

class Task(threading.Thread):


    def __init__(self, *args, **kwargs):
        """ 
            __init__()

            thread that runs a function.  

        """
        threading.Thread.__init__(self, *args, **kwargs)
        self._queue = queue.Queue()
        self.setDaemon(True)
        self.result = None
        self.once = True
        
    def __iter__(self):
        for x in dir(self): yield x

    ## INPUT

    def put(self, *args, **kwargs):
        """ 
            put()

            put args, kwargs to the task.

        """
        self._last = time.time()
        self._queue.put((args, kwargs))
        return self

    def join(self, *args, **kwargs):
        """
             join()

             join a Task. 

        """
        threading.Thread.join(self, *args, **kwargs)
        return self.result

    ## RUNNING

    def run(self, *args, **kwargs):
        """
            run()

            runs the function.

        """
        from meds.kernel import kernel
        self.type = "Task"
        self._start = time.time()
        self._last = time.time()
        self._status = "ok"
        self._name = ""
        self._event = None
        while self._status:
            arglist, kwargdict = self._queue.get()
            if not arglist or not arglist[0]: break
            self.func, *arguments = arglist
            self._name = named(self.func)
            self.setName(self._name)
            if len(arguments) > 0: self._event = arguments[0]
            self._status = "running"
            self.result = self.func(*arguments, **kwargdict)
            self._last = time.time()
            if "once" in self: self._status = "break" ; break
            else: self._status = "done"
            logging.info("! Task/%s %s %.6ss" % (self._name, self._status, self._last - self._start))
        return self.result

    def exit(self):
        """
            exit()

            quits the task. 

        """
        self._status = ""
        self.put(None, {})

## DISPATCHER

class Dispatcher(Object):

    cc = "!"
    default = ""
    threads = []
    handlers = Object()

    def __init__(self, *args, **kwargs):
        """
            __init__()

            delegates workload to threads. threads gets instantiated when needed.
        
        """
        Object.__init__(self, *args, **kwargs)
        self._input = queue.Queue()
        self._output = queue.Queue()
        self._ready = threading.Event()
        self._status = "start"

    ## VIRTUAL

    def dispatch(self, *args, **kwargs): 
        """
            dispatch()

            virtual method.

        """
        pass

    ## INPUT

    def start(self, *args, **kwargs):
        """
            start(func)

            start a task. use args and kwargs to pass on arguments.

        """
        task = Task()
        task.start()
        task.put(*args, **kwargs)
        return task

    ## RESULTS

    def collect(self, thrs):
        """
             collect(thrs)

             join all threads given as arguments.

        """
        result = []
        for thr in thrs:
            if thr:
                res = thr.join()
                if res: result.append(res)
        return result

    ## STOP

    def stop(self):
        """
            stop()

            stop the dispatcher. 

        """
        self._status = ""
        self.start(None)

    def register(self, cmnd, func):
        reg(self, cmnd, func)

    def shutdown(self, name=""):
        """
            shutdown()

            stop all running threads.

        """
        for thr in threading.enumerate():
            if str(thr).startswith("<_"): continue
            if name and name not in str(thr): continue
            if "stop" in dir(thr): thr.stop()
            elif "exit" in dir(thr): thr.exit()
            elif "cancel" in dir(thr): thr.cancel()
            else: thr.join(0.1)
            logging.warn("# shutdown %s" % naming(thr))
        return True
