# kernel.py
#
#

""" runtime namespace. """

__copyright__ = "Copyright 2015, Bart Thate"

## IMPORTS

from meds.utils.opts import make_opts, opts_defs
from meds.dispatcher import Dispatcher
from meds.utils.other import hello, show_eggs
from meds.utils.misc import match, reg
from meds.utils.trace import get_plugname
from meds.errors import error, IsMethod
from meds.defines import GREEN, ENDC
from meds.object import Object, OO
from meds.parsed import Parsed
from meds.log import log_init

import importlib
import inspect
import logging
import pkgutil
import types
import time
import sys
import os

## CONTROL CHARACTERS

cc = Object()
cc.Bot = "none"
cc.TestBot = "!"
cc.CLI = "none"
cc.IRC = "!"
cc.XMPP = "none"

## KERNEL

class Kernel(Dispatcher):

    """ global namespace of the bot, named Kernel. """

    def __init__(self, *args, **kwargs):
        Dispatcher.__init__(self, *args, **kwargs)
        self.cmnds = Object()
        self.fleet = OO()
        self.plugnames = []

    def announce(self, txt):
        """ announce a text on all channels of all bots. """
        thrs = []
        for bot in self.fleet: thrs.append(self.start(bot.announce, txt))
        if thrs: self.collect(thrs)
        return thrs

    def register(self, cmnd, func): 
        """ register a command. """
        reg(self.cmnds, cmnd, func)

    def boot(self, *args, **kwargs):
        """ enable runtime services. """
        from meds.cfg import cfg
        global fleet
        self._start = time.time()
        opts, arglist = make_opts(opts_defs)
        cfg.kernel.update(vars(opts))
        for key,value in kwargs.items():
            if not getattr(cfg.kernel , key, ""): cfg.kernel[key] = value
        if not cfg.kernel.workdir: cfg.kernel.workdir = os.path.expanduser("~/.meds")
        cfg.kernel.runargs = arglist
        log_init(cfg.kernel.loglevel, colors=cfg.kernel.colors)
        if cfg.kernel.shell: hello(*args, **cfg.kernel)
        if cfg.kernel.path:
            for path in sys.path: print(path)
            print()
        for package in cfg.kernel.packages: self.load(package)
        for mod in cfg.kernel.modules.split(","):
            if mod: kernel.load(mod)
        self.init(cfg.kernel.init)
        self.ready()

    def init(self, plugnames):
        """ initialize plugins with the kernel. """
        modnames = plugnames.split(",") 
        for modname in sorted(modnames, reverse=True):
            if not modname: continue
            objs = []
            for name, object in self.items():
                if modname in name: objs.append(object)
            for obj in objs:
                logging.warn("# %s/init %s" % (self.type, modname))
                if modname == "xmpp": obj.init() ; continue
                init = getattr(obj, "init", None)
                if init: init()
        self.ready()

    ## DISPATCH

    def dispatch(self, *args, **kwargs):
        """ dispatch an event. """
        p = Parsed(*args, **kwargs)
        cmnds = match(self.cmnds, p.cmnd)
        thrs = []
        for cmnd in cmnds: 
            try: o = self.start(cmnd.func, p) ; thrs.append(o)
            except AttributeError: continue
            except: error()
        return thrs

    ## LOAD

    def load(self, name):
        """ load plugins from a package. """
        from meds.cfg import cfg
        self[name] = mod = self.loading(name)
        if "__path__" in dir(mod):
            for pkg in pkgutil.walk_packages(mod.__path__, name + "."):
                n = pkg[1]
                do = True
                for mod in cfg.kernel.exclude:
                    if mod in n: do = False
                if not do: continue
                logging.info("load %s" % n)
                self[n] = self.loading(n)
                register = getattr(self[n], "register", None)
                if register: register()
        return self[name]

    def loading(self, name, package=None, force=True):
        """ load module. """
        if name in self and not force: mod = self.reload(name)
        else: mod = importlib.import_module(name, package)
        return mod
 
    ## RELOAD

    def reload(self, name):
        """ reload a module. """
        logging.warn("# %s/reload %s" % (self.type, name))
        self[name] = importlib.reload(self[name])
        return self[name]

    def reloading(self, name):
        """ reload all modules matching. """
        names = search(self, name)
        for name in names: self.reload(name)

## DEFINES

kernel = Kernel()
