# object.py
#
#

""" core object. A dict with dotted named access. """

__copyright__ = "Copyright 2015, Bart Thate"

## IMPORTS

from meds.errors import NoJSON, BotNotSet, AttributeSet
from meds.utils.file import list_files
from meds.utils.time import fn_time, rtime
from meds.utils.name import typed, name
from meds.utils.lock import locked
from meds.utils.misc import urled, root, signature, dumps, slice
from meds.utils.misc import path as dopath
import threading
import hashlib
import logging
import inspect
import errno
import fcntl
import types
import json
import time
import os

## OBJECT OF OBJECTS

class OO(list): pass

## OBJECT

class Object(dict):

    """ object on which the rest of the program is based. """

    def __getattribute__(self, name):
        try: val = self[name]
        except KeyError: val = super().__getattribute__(name)
        return val

    def __getattr__(self, name):
        if name == "url": return urled(self)
        if name == "_ready":  self._ready = threading.Event()
        if name == "skip": return ["parsed", ] 
        if name == "cc": return "!"
        if name == "type": return typed(self)
        if name not in self: raise AttributeError(name)
        return self[name]

    def __setitem__(self, name, value):
        return dict.__setitem__(self, name, value)

    def __setattr__(self, name, value):
        try:
            val = super().__getattribute__(name)
            if inspect.ismethod(val): raise AttributeSet        
        except: pass
        return dict.__setitem__(self, name, value)

    def __contains__(self, name):    
        try: self[name] ; return True
        except KeyError: return False

    ## WAITERS

    def ready(self):
        """ signal to ready state. """
        logging.debug("! %s/ready" % self.type)
        self._ready.set()

    def clear(self):
        """ clear the ready state. """
        logging.debug("! %s/clean" % self.type)
        self._ready.clear()

    def wait(self, sec=180.0):
        """ wait for ready state. """
        logging.debug("! %s/wait" % self.type)
        self._ready.wait(sec)

    ## READ

    def load(self, path=""):
        """ load a JSON file. """
        if not path: path = self._path
        ondisk = self.read(path)
        fromdisk = json.loads(ondisk)
        if "data" in fromdisk: self.update(fromdisk["data"])
        else: self.update(fromdisk)
        if "saved" in fromdisk: self.saved = fromdisk["saved"]
        self._path = path
        return self

    def read(self, path):
        """ read a JSON file. """
        try: f = open(path, "r", encoding="utf-8")
        except IOError as ex:
            if ex.errno == errno.ENOENT: return "{}"
            raise
        res = ""
        for line in f:
            if not line.strip().startswith("#"): res += line
        if not res.strip(): return "{}"
        f.close()
        return res

    ## WRITE

    def prepare(self):
        """ create JSON string. """
        from meds.utils.trace import get_plugname
        todisk = Object()
        todisk.data = slice(self)
        todisk.saved_from = get_plugname(2)
        todisk.type = self.type
        todisk.saved = self.saved = time.ctime(time.time())
        todisk.signature = signature(todisk.data)
        try: result = dumps(todisk, indent=4, ensure_ascii=False, sort_keys=True)
        except TypeError: raise NoJSON()
        return result

    def save(self, stime=""):
        """ save to disk. """
        from meds.utils.time import rtime
        from meds.utils.join import j
        if not stime: stime = rtime()
        self.sync(j(dopath(self), stime))
        return stime

    @locked
    def sync(self, path="", **kwargs):
        """ 
            sync(path="")

            sync to disk. 

        """
        from meds.utils.other import headertxt
        from meds.utils.time import rtime
        from meds.utils.file import cdir
        from meds.utils.join import j
        if not path:
            try: path = self._path
            except AttributeError: path = self._path = j(dopath(self), rtime())
        logging.info("! sync %s" % path)
        d, fn = os.path.split(path)
        cdir(d)
        todisk = self.prepare(**kwargs)
        datafile = open(os.path.abspath(path) + ".tmp", 'w')
        fcntl.flock(datafile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        datafile.write(headertxt % "%s characters" % len(todisk))
        datafile.write(todisk)
        datafile.write("\n")
        fcntl.flock(datafile, fcntl.LOCK_UN)
        datafile.close()
        os.rename(path + ".tmp", path)
        return path
