# store.py
#
#

""" read files from disk. """

__copyright__ = "Copyright 2015, Bart Thate"

## IMPORTS

from meds.utils.misc import path, timed
from meds.utils.time import fn_time
from meds.parsed import Parsed
from meds.object import Object
from meds.kernel import kernel
from meds.cfg import cfg

import logging
import glob
import os

## STORE

class Store(Object):

    ## ITERATORS

    def list(self, path, *args, **kwargs):
        """
            list(path, prefix, time)

            list all files matching path, prefix or time.

        """
        p = Parsed(kwargs)
        if args: prefix = args[0]
        else: prefix = ""
        res = []
        if not path.endswith(os.sep): path += os.sep
        if prefix and os.path.isdir(path + prefix): path += prefix 
        if "*" not in path: path += "*"
        logging.info("path %s" % path)
        for fnn in glob.glob(path):
            if not prefix and p.time not in fnn: continue
            if os.path.isdir(fnn): res.extend(self.list(fnn, **kwargs)) ; continue
            else: res.append(fnn)
        return res 

    def all(self, *args, **kwargs):
        """
             all(prefix, time)

             select all records based on a optional prefix, counter start and time 

        """
        return sorted(self.list(path(self), *args, **kwargs), key=lambda x: fn_time(x))

    def selected(self, event):
        """
             selected(self, parsed)

             select all records based on a parsed object.

        """
        from meds.utils.misc import path, selector, wanted, notwanted
        objs = []
        uniqlist = []
        for fn in self.all(*event.args, **event.want):
            try: obj = Object().load(fn)
            except: logging.warn("fail %s" % fn) ; continue
            if "deleted" in obj and obj.deleted: continue
            if not selector(obj, event.args): continue
            if not wanted(obj, event.want): continue
            if notwanted(obj, event.notwant): continue
            uniq = event.switch.get("uniq", "")
            if uniq:
                val = obj.get(uniq, "")
                if val in uniqlist: continue
                else: uniqlist.append(val)
            objs.append(obj)
        return objs

    def prefixes(self, *args, **kwargs):
        """
            prefixes(path, prefix, time)

            show prefixes of directory.

        """
        from meds.cfg import cfg
        fns = self.all(*args, **kwargs)
        prefixes = []
        for fn in fns:
            workdir, path = fn.split(cfg.kernel.workdir) 
            boh, prefix, *rest = path.split(os.sep)
            if prefix not in prefixes: prefixes.append(prefix)
        return prefixes

    def first(self, *args, **kwargs):
        """ 
            first(path, prefix, time)

            find first object.

        """
        fns = self.all(*args, **kwargs)
        for fn in fns:
            obj = Object().load(fn)
            if "deleted" in obj and obj.deleted: continue
            if args and len(args) > 1 and args[1] != obj.get(args[0], ""): continue
            return obj

    def last(self, *args, **kwargs):
        """
            last(path, prefix, time)

            find last object.

        """
        fns = self.all(*args, **kwargs)[::-1]
        res = None
        for fn in fns[::-1]:
            obj = Object().load(fn)
            if "deleted" in obj and obj.deleted: continue
            if args and len(args) > 1 and args[1] != obj.get(args[0], ""): continue
            res = obj
        return res

## CMNDS

def rm(event):
    """ rm <key> <match> - remove objects """        
    try: key, match = event.rest.split(" ", 1)
    except ValueError: return
    nr = 0
    store = Store()
    for fn in store.all(key):
        obj = Object().load(fn)
        if key not in obj: continue
        if match not in obj[key]: continue
        obj.deleted = True
        obj.sync()
        nr += 1
    event.ok(nr)

def restore(*args, **kwargs):
    """ restore <key> <match> """
    p = Parsed(*args, **kwargs)
    try: key, match = p.args
    except ValueError: return
    nr = 0
    store = Store()
    for fn in store.all(*p.args, **p.kwargs):
        obj = Object().load(fn)
        if key not in obj: continue
        if match not in obj[key]: continue
        if "deleted" not in obj: continue
        obj.deleted = False
        obj.sync()
        nr += 1
    event.ok(nr)

def find(event):
    """ find <key>"""
    nr = 1
    store = Store()
    objs = sorted(store.selected(event), key=lambda x: timed(x))
    for obj in objs: event.display(obj, event.args, str(nr)) ; nr += 1
      
def dump(event):
    if not event.args: return
    store = Store()
    objs = sorted(store.selected(event), key=lambda x: x.timed())
    for obj in objs: event.reply(obj.json(indent=4))

def fields(event):
    if not event.args: return
    store = Store()
    search_term = event.args[0]
    keys = []
    nr = 0
    for fn in store.all():
        obj = Object().load(fn)
        for key in obj:
            if search_term in str(obj[key]) and key not in keys: keys.append(key) ; nr += 1
    if keys: event.reply(", ".join(keys))

def deleted(event):
    if not event.args: return
    store = Store()
    key = event.args[0]
    nr = 0
    for fn in store.all(key):
       obj = Object().load(fn) 
       if "deleted" not in obj: continue
       event.reply(obj.get(key, ""))
       nr += 1

def first(event):
    store = Store()
    res = store.first(*event.args)
    if res: event.display(res, ["saved"] + list(event.args))

def last(event):
    store = Store()
    res = store.last(*event.args)
    if res: event.display(res, ["saved"] + list(event.args))

def mbox(event):
    if not event.args: return
    fn = os.path.expanduser(event.args[0])
    nr = 0
    try: object = mailbox.Maildir(fn, create=False)
    except: 
        try: object = mailbox.mbox(fn, create=False)
        except: event.reply("need a mbox or maildir.") ; return
    for m in object:
        o = Object()
        o.update(m.items())
        try: sdate = os.sep.join(short_date(o.Date).split())
        except AttributeError: sdate = None
        o.text = bytes()
        for load in m.walk():
            if load.get_content_type() == 'text/plain': o.text += load.get_payload(decode=True)
        o.text = str(o.text, "latin-1")
        o.prefix = "email"
        if sdate: o.save(sdate)
        else: o.save()
        nr += 1
    event.ok(nr)
    return nr

def log(event):
    o = Object(event)
    o.prefix = "log"
    o.log = event.rest
    path = o.save()
    event.ok(1)

def todo(event):
    o = Object(event)
    o.prefix = "todo"
    o.todo = event.rest
    path = o.save()
    event.ok(1)

def status(event):
    o = Object(event)
    o.prefix = "status"
    o.status = event.rest
    path = o.save()
    event.ok(1)

def user(event):
    if cfg.kernel.owner != event.origin: event.reply("ENOTALLOWED %s" % event.origin) ; return
    if "@" not in event.rest: event.reply("ENOTAUSERHOST %s" % event.rest) ; return
    o = Object(event)
    o.prefix = "user"
    o.user = event.rest
    path = o.save()
    event.ok(1)

def rss(event):
    o = Object(event)
    o.prefix = "rss"
    o.rss = event.rest  
    o.service = "rss"
    path = o.save()  
    event.ok(1)

def symptom(event):
    o = Object(event)
    o.prefix = "symptom"
    o.symptom = event.rest 
    if "karma" in event: o.karma = event.karma 
    path = o.save()  
    event.ok(1)

def register():
    kernel.register("deleted", deleted)
    kernel.register("dump", dump)
    kernel.register("fields", fields)
    kernel.register("find", find)
    kernel.register("first", first)
    kernel.register("last", last)
    kernel.register("mbox", mbox)
    kernel.register("log", log)
    kernel.register("rm", rm)
    kernel.register("restore", restore)
    kernel.register("rss", rss)   
    kernel.register("status", status)
    kernel.register("todo", todo)
    kernel.register("symptom", symptom)
    kernel.register("user", user)
